@@ -2,7 +2,7 @@
 #
 #    BitcoinLib - Python Cryptocurrency Library
 #    BLOCK parsing and construction
-#    © 2020 Juni - 1200 Web Development <http://1200wd.com/>
+#    © 2020 June - 1200 Web Development <http://1200wd.com/>
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU Affero General Public License as
@@ -18,9 +18,10 @@
 #    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 
+from io import BytesIO
 from bitcoinlib.encoding import *
 from bitcoinlib.networks import Network
-from bitcoinlib.transactions import transaction_deserialize, Transaction
+from bitcoinlib.transactions import Transaction
 
 
 class Block:
@@ -32,7 +33,7 @@ class Block:
 
         >>> b = Block('0000000000000000000154ba9d02ddd6cee0d71d1ea232753e02c9ac6affd709', version=0x20000000, prev_block='0000000000000000000f9578cda278ae7a2002e50d8e6079d11e2ea1f672b483', merkle_root='20e86f03c24c53c12014264d0e405e014e15a02ad02c174f017ee040750f8d9d', time=1592848036, bits=387044594, nonce=791719079)
         >>> b
-        <Block(0000000000000000000154ba9d02ddd6cee0d71d1ea232753e02c9ac6affd709, None, transactions: None)>
+        <Block(0000000000000000000154ba9d02ddd6cee0d71d1ea232753e02c9ac6affd709, None, transactions: 0)>
 
         :param block_hash: Hash value of serialized block
         :type block_hash: bytes, str
@@ -60,35 +61,37 @@ class Block:
 
         self.block_hash = to_bytes(block_hash)
         if isinstance(version, int):
-            self.version = struct.pack('>L', version)
+            self.version = version.to_bytes(4, byteorder='big')
             self.version_int = version
         else:
             self.version = to_bytes(version)
-            self.version_int = 0 if not self.version else struct.unpack('>L', self.version)[0]
+            self.version_int = 0 if not self.version else int.from_bytes(self.version, 'big')
         self.prev_block = to_bytes(prev_block)
         self.merkle_root = to_bytes(merkle_root)
         self.time = time
         if not isinstance(time, int):
-            self.time = struct.unpack('>L', time)[0]
+            self.time = int.from_bytes(time, 'big')
         if isinstance(bits, int):
-            self.bits = struct.pack('>L', bits)
+            self.bits = bits.to_bytes(4, 'big')
             self.bits_int = bits
         else:
             self.bits = to_bytes(bits)
-            self.bits_int = 0 if not self.bits else struct.unpack('>L', self.bits)[0]
+            self.bits_int = 0 if not self.bits else int.from_bytes(self.bits, 'big')
         if isinstance(nonce, int):
-            self.nonce = struct.pack('>L', nonce)
+            self.nonce = nonce.to_bytes(4, 'big')
             self.nonce_int = nonce
         else:
             self.nonce = to_bytes(nonce)
-            self.nonce_int = 0 if not self.nonce else struct.unpack('>L', self.nonce)[0]
+            self.nonce_int = 0 if not self.nonce else int.from_bytes(self.nonce, 'big')
         self.transactions = transactions
+        if self.transactions is None:
+            self.transactions = []
         self.txs_data = None
         self.confirmations = confirmations
         self.network = network
         if not isinstance(network, Network):
             self.network = Network(network)
-        self.tx_count = None
+        self.tx_count = 0
         self.page = 1
         self.limit = 0
         self.height = height
@@ -96,10 +99,10 @@ class Block:
                 and self.version_int > 1:
             # first bytes of unlocking script of coinbase transaction contains block height (BIP0034)
             if self.transactions[0].coinbase and self.transactions[0].inputs[0].unlocking_script:
-                calc_height = struct.unpack('<L', self.transactions[0].inputs[0].unlocking_script[1:4] + b'\x00')[0]
-                if height and calc_height != height:
-                    raise ValueError("Specified block height is different than calculated block height according to "
-                                     "BIP0034")
+                calc_height = int.from_bytes(self.transactions[0].inputs[0].unlocking_script[1:4] + b'\x00', 'little')
+                if height and calc_height != height and height > 227835:
+                    raise ValueError("Specified block height %d is different than calculated block height according to "
+                                     "BIP0034" % height)
                 self.height = calc_height
 
     def check_proof_of_work(self):
@@ -116,14 +119,151 @@ class Block:
         """
         if not self.block_hash or not self.bits:
             return False
-        if change_base(self.block_hash, 256, 10) < self.target:
+        if int.from_bytes(self.block_hash, 'big') < self.target:
             return True
         return False
 
     def __repr__(self):
-        return "<Block(%s, %s, transactions: %s)>" % (to_hexstring(self.block_hash), self.height, self.tx_count)
+        return "<Block(%s, %s, transactions: %s)>" % (self.block_hash.hex(), self.height, self.tx_count)
 
     @classmethod
+    def parse(cls, raw, block_hash=None, height=None, parse_transactions=False, limit=0, network=DEFAULT_NETWORK):
+        """
+        Create Block object from raw serialized block in bytes or BytesIO format. Wrapper for :func:`parse_bytesio`
+
+        Get genesis block:
+
+        >>> from bitcoinlib.services.services import Service
+        >>> srv = Service()
+        >>> b = srv.getblock(0)
+        >>> b.block_hash.hex()
+        '000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f'
+
+        :param raw: Raw serialize block
+        :type raw: BytesIO, bytes
+        :param block_hash: Specify block hash if known to verify raw block. Value error will be raised if calculated block hash is different than specified.
+        :type block_hash: bytes
+        :param height: Specify height if known. Will be derived from coinbase transaction if not provided.
+        :type height: int
+        :param parse_transactions: Indicate if transactions in raw block need to be parsed and converted to Transaction objects. Default is False
+        :type parse_transactions: bool
+        :param limit: Maximum number of transactions to parse. Default is 0: parse all transactions. Only used if parse_transaction is set to True
+        :type limit: int
+        :param network: Name of network
+        :type network: str
+
+        :return Block:
+        """
+
+        if isinstance(raw, bytes):
+            return cls.parse_bytesio(BytesIO(raw), block_hash, height, parse_transactions, limit, network)
+        else:
+            return cls.parse_bytesio(raw, block_hash, height, parse_transactions, limit, network)
+
+    @classmethod
+    def parse_bytes(cls, raw_bytes, block_hash=None, height=None, parse_transactions=False, limit=0,
+                    network=DEFAULT_NETWORK):
+        """
+        Create Block object from raw serialized block in bytes or BytesIO format. Wrapper for :func:`parse_bytesio`
+
+        Get genesis block:
+
+        >>> from bitcoinlib.services.services import Service
+        >>> srv = Service()
+        >>> b = srv.getblock(0)
+        >>> b.block_hash.hex()
+        '000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f'
+
+        :param raw_bytes: Raw serialize block
+        :type raw_bytes: bytes
+        :param block_hash: Specify block hash if known to verify raw block. Value error will be raised if calculated block hash is different than specified.
+        :type block_hash: bytes
+        :param height: Specify height if known. Will be derived from coinbase transaction if not provided.
+        :type height: int
+        :param parse_transactions: Indicate if transactions in raw block need to be parsed and converted to Transaction objects. Default is False
+        :type parse_transactions: bool
+        :param limit: Maximum number of transactions to parse. Default is 0: parse all transactions. Only used if parse_transaction is set to True
+        :type limit: int
+        :param network: Name of network
+        :type network: str
+
+        :return Block:
+        """
+
+        raw_bytesio = BytesIO(raw_bytes)
+        return cls.parse_bytesio(raw_bytesio, block_hash, height, parse_transactions, limit, network)
+
+    @classmethod
+    def parse_bytesio(cls, raw, block_hash=None, height=None, parse_transactions=False, limit=0,
+                      network=DEFAULT_NETWORK):
+        """
+        Create Block object from raw serialized block in BytesIO format
+
+        Get genesis block:
+
+        >>> from bitcoinlib.services.services import Service
+        >>> srv = Service()
+        >>> b = srv.getblock(0)
+        >>> b.block_hash.hex()
+        '000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f'
+
+        :param raw: Raw serialize block
+        :type raw: BytesIO
+        :param block_hash: Specify block hash if known to verify raw block. Value error will be raised if calculated block hash is different than specified.
+        :type block_hash: bytes
+        :param height: Specify height if known. Will be derived from coinbase transaction if not provided.
+        :type height: int
+        :param parse_transactions: Indicate if transactions in raw block need to be parsed and converted to Transaction objects. Default is False
+        :type parse_transactions: bool
+        :param limit: Maximum number of transactions to parse. Default is 0: parse all transactions. Only used if parse_transaction is set to True
+        :type limit: int
+        :param network: Name of network
+        :type network: str
+
+        :return Block:
+        """
+        block_header = raw.read(80)
+        block_hash_calc = double_sha256(block_header)[::-1]
+        if not block_hash:
+            block_hash = block_hash_calc
+        elif block_hash != block_hash_calc:
+            raise ValueError("Provided block hash does not correspond to calculated block hash %s" %
+                             block_hash_calc.hex())
+
+        raw.seek(0)
+        version = raw.read(4)[::-1]
+        prev_block = raw.read(32)[::-1]
+        merkle_root = raw.read(32)[::-1]
+        time = raw.read(4)[::-1]
+        bits = raw.read(4)[::-1]
+        nonce = raw.read(4)[::-1]
+        tx_count = read_varbyteint(raw)
+        tx_start_pos = raw.tell()
+        txs_data_size = raw.seek(0, 2)
+        raw.seek(tx_start_pos)
+        transactions = []
+
+        while parse_transactions and raw.tell() < txs_data_size:
+            if limit != 0 and len(transactions) >= limit:
+                break
+            t = Transaction.parse_bytesio(raw, strict=False)
+            transactions.append(t)
+            # TODO: verify transactions, need input value from previous txs
+            # if verify and not t.verify():
+            #     raise ValueError("Could not verify transaction %s in block %s" % (t.txid, block_hash))
+
+        if parse_transactions and limit == 0 and tx_count != len(transactions):
+            raise ValueError("Number of found transactions %d is not equal to expected number %d" %
+                             (len(transactions), tx_count))
+
+        block = cls(block_hash, version, prev_block, merkle_root, time, bits, nonce, transactions, height,
+                    network=network)
+        block.txs_data = raw
+        block.tx_count = tx_count
+        return block
+
+    @classmethod
+    @deprecated
     def from_raw(cls, raw, block_hash=None, height=None, parse_transactions=False, limit=0, network=DEFAULT_NETWORK):
         """
         Create Block object from raw serialized block in bytes.
@@ -156,7 +296,7 @@ class Block:
             block_hash = block_hash_calc
         elif block_hash != block_hash_calc:
             raise ValueError("Provided block hash does not correspond to calculated block hash %s" %
-                             to_hexstring(block_hash_calc))
+                             block_hash_calc.hex())
 
         version = raw[0:4][::-1]
         prev_block = raw[4:36][::-1]
@@ -165,26 +305,24 @@ class Block:
         bits = raw[72:76][::-1]
         nonce = raw[76:80][::-1]
         tx_count, size = varbyteint_to_int(raw[80:89])
-        txs_data = raw[80+size:]
+        txs_data = BytesIO(raw[80+size:])
 
         # Parse coinbase transaction so we can extract extra information
-        transactions = [transaction_deserialize(txs_data, network=network, check_size=False)]
-        txs_data = txs_data[transactions[0].size:]
+        # transactions = [Transaction.parse(txs_data, network=network)]
+        # txs_data = BytesIO(txs_data[transactions[0].size:])
+        # block_txs_data = txs_data.read()
+        txs_data_size = txs_data.seek(0, 2)
+        txs_data.seek(0)
+        transactions = []
 
-        while parse_transactions and txs_data:
+        while parse_transactions and txs_data and txs_data.tell() < txs_data_size:
             if limit != 0 and len(transactions) >= limit:
                 break
-            t = transaction_deserialize(txs_data, network=network, check_size=False)
+            t = Transaction.parse_bytesio(txs_data, strict=False)
             transactions.append(t)
-
-            # t.rawtx = b''
-            # traw = t.raw_hex()
-            # tblock = txs_data[:t.size].hex()
-            # if traw != tblock:
-            #     print(t.txid)
-            #     print(traw)
-            #     print(tblock)
-            txs_data = txs_data[t.size:]
+            # t = transaction_deserialize(txs_data, network=network, check_size=False)
+            # transactions.append(t)
+            # txs_data = txs_data[t.size:]
             # TODO: verify transactions, need input value from previous txs
             # if verify and not t.verify():
             #     raise ValueError("Could not verify transaction %s in block %s" % (t.txid, block_hash))
@@ -209,10 +347,9 @@ class Block:
         :return:
         """
         n = 0
-        while self.txs_data and (limit == 0 or n < limit):
-            t = transaction_deserialize(self.txs_data, network=self.network, check_size=False)
+        while self.txs_data and (limit == 0 or n < limit) and len(self.transactions) < self.tx_count:
+            t = Transaction.parse_bytesio(self.txs_data, strict=False, network=self.network)  # , check_size=False
             self.transactions.append(t)
-            self.txs_data = self.txs_data[t.size:]
             n += 1
 
     def as_dict(self):
@@ -222,11 +359,11 @@ class Block:
         :return dict:
         """
         return {
-            'block_hash': to_hexstring(self.block_hash),
+            'block_hash': self.block_hash.hex(),
             'height': self.height,
             'version': self.version_int,
-            'prev_block': None if not self.prev_block else to_hexstring(self.prev_block),
-            'merkle_root': to_hexstring(self.merkle_root),
+            'prev_block': None if not self.prev_block else self.prev_block.hex(),
+            'merkle_root': self.merkle_root.hex(),
             'timestamp': self.time,
             'bits': self.bits_int,
             'nonce': self.nonce_int,
@@ -248,9 +385,7 @@ class Block:
         if not self.bits:
             return 0
         exponent = self.bits[0]
-        if not PY3:
-            exponent = ord(exponent)
-        coefficient = struct.unpack('>L', b'\x00' + self.bits[1:])[0]
+        coefficient = int.from_bytes(b'\x00' + self.bits[1:], 'big')
         return coefficient * 256 ** (exponent - 3)
 
     @property
@@ -262,7 +397,7 @@ class Block:
         """
         if not self.bits:
             return ''
-        return hex(self.target)[2:].zfill(64)
+        return hex(int(self.target))[2:].zfill(64)
 
     @property
     def difficulty(self):
@@ -306,7 +441,7 @@ class Block:
         rb = self.version[::-1]
         rb += self.prev_block[::-1]
         rb += self.merkle_root[::-1]
-        rb += struct.pack('>L', self.time)[::-1]
+        rb += self.time.to_bytes(4, 'little')
         rb += self.bits[::-1]
         rb += self.nonce[::-1]
         if len(rb) != 80:
