@@ -2,7 +2,7 @@
 #
 #    BitcoinLib - Python Cryptocurrency Library
 #    DataBase - SqlAlchemy database definitions
-#    © 2016 - 2020 February - 1200 Web Development <http://1200wd.com/>
+#    © 2016 - 2020 October - 1200 Web Development <http://1200wd.com/>
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU Affero General Public License as
@@ -18,65 +18,85 @@
 #    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 
-try:
-    import enum
-except ImportError:
-    import enum34 as enum
 from sqlalchemy import create_engine
 from sqlalchemy import (Column, Integer, BigInteger, UniqueConstraint, CheckConstraint, String, Boolean, Sequence,
-                        ForeignKey, DateTime, Numeric, Text)
+                        ForeignKey, DateTime, LargeBinary)
 from sqlalchemy.ext.declarative import declarative_base
-from sqlalchemy.orm import sessionmaker, relationship
-try:
-    from urllib.parse import urlparse
-except ImportError:
-    from urlparse import urlparse
+from sqlalchemy.ext.compiler import compiles
+from sqlalchemy.orm import sessionmaker, relationship, close_all_sessions
+from urllib.parse import urlparse
 from bitcoinlib.main import *
 
 _logger = logging.getLogger(__name__)
-_logger.info("Using Database %s" % DEFAULT_DATABASE)
 Base = declarative_base()
 
 
-class DbInit:
+@compiles(LargeBinary, "mysql")
+def compile_largebinary_mysql(type_, compiler, **kwargs):
+    length = type_.length
+    element = "BLOB" if not length else "VARBINARY(%d)" % length
+    return element
+
+
+class Db:
     """
-    Initialize database and open session
+    Bitcoinlib Database object used by Service() and HDWallet() class. Initialize database and open session when
+    creating database object.
 
     Create new database if is doesn't exist yet
 
     """
-    def __init__(self, db_uri=None):
+    def __init__(self, db_uri=None, password=None):
         if db_uri is None:
             db_uri = DEFAULT_DATABASE
-        o = urlparse(db_uri)
-        if not o.scheme or \
-                len(o.scheme) < 2:  # Dirty hack to avoid issues with urlparse on Windows confusing drive with scheme
-            db_uri = 'sqlite:///%s' % db_uri
-        if db_uri.startswith("sqlite://") and ALLOW_DATABASE_THREADS:
-            if "?" in db_uri: db_uri += "&"
-            else: db_uri += "?"
-            db_uri += "check_same_thread=False"
-        self.engine = create_engine(db_uri, isolation_level='READ UNCOMMITTED')
-        Session = sessionmaker(bind=self.engine)
+        self.o = urlparse(db_uri)
+        if not self.o.scheme or \
+                len(self.o.scheme) < 2:  # Dirty hack to avoid issues with urlparse on Windows confusing drive with scheme
+            if password:
+                # Warning: This requires the pysqlcipher3 module
+                db_uri = 'sqlite+pysqlcipher://:%s@/%s?cipher=aes-256-cfb&kdf_iter=64000' % (password, db_uri)
+            else:
+                db_uri = 'sqlite:///%s' % db_uri
 
+        if db_uri.startswith("sqlite://") and ALLOW_DATABASE_THREADS:
+            db_uri += "&" if "?" in db_uri else "?"
+            db_uri += "check_same_thread=False"
+        if self.o.scheme == 'mysql':
+            db_uri += "&" if "?" in db_uri else "?"
+            db_uri += 'binary_prefix=true'
+        self.engine = create_engine(db_uri, isolation_level='READ UNCOMMITTED')
+
+        Session = sessionmaker(bind=self.engine)
         Base.metadata.create_all(self.engine)
         self._import_config_data(Session)
-
         self.session = Session()
 
+        _logger.info("Using database: %s://%s:%s/%s" % (self.o.scheme or '', self.o.hostname or '',
+                                                        self.o.port or '', self.o.path or ''))
+        self.db_uri = db_uri
+
         # VERIFY AND UPDATE DATABASE
         # Just a very simple database update script, without any external libraries for now
         #
+        version_db = self.session.query(DbConfig.value).filter_by(variable='version').scalar()
+        if version_db[:3] == '0.4' and BITCOINLIB_VERSION[:3] == '0.5':
+            raise ValueError("Old database version found (<0.4.19). Can not connect to 0.5 version database "
+                             "automatically, use db_update tool to update")
         try:
-            version_db = self.session.query(DbConfig.value).filter_by(variable='version').scalar()
             if BITCOINLIB_VERSION != version_db:
                 _logger.warning("BitcoinLib database (%s) is from different version then library code (%s). "
                                 "Let's try to update database." % (version_db, BITCOINLIB_VERSION))
                 db_update(self, version_db, BITCOINLIB_VERSION)
-
         except Exception as e:
             _logger.warning("Error when verifying version or updating database: %s" % e)
 
+    def drop_db(self, yes_i_am_sure=False):
+        if yes_i_am_sure:
+            self.session.commit()
+            self.session.close_all()
+            close_all_sessions()
+            Base.metadata.drop_all(self.engine)
+
     @staticmethod
     def _import_config_data(ses):
         session = ses()
@@ -105,7 +125,7 @@ def add_column(engine, table_name, column):
     """
     column_name = column.compile(dialect=engine.dialect)
     column_type = column.type.compile(engine.dialect)
-    engine.execute('ALTER TABLE %s ADD COLUMN %s %s' % (table_name, column_name, column_type))
+    engine.execute("ALTER TABLE %s ADD COLUMN %s %s" % (table_name, column_name, column_type))
 
 
 class DbConfig(Base):
@@ -203,8 +223,8 @@ class DbKey(Base):
                        "depth=1 are the masterkeys children.")
     change = Column(Integer, doc="Change or normal address: Normal=0, Change=1")
     address_index = Column(BigInteger, doc="Index of address in HD key structure address level")
-    public = Column(String(512), index=True, doc="Hexadecimal representation of public key")
-    private = Column(String(512), index=True, doc="Hexadecimal representation of private key")
+    public = Column(LargeBinary(128), index=True, doc="Bytes representation of public key")
+    private = Column(LargeBinary(128), index=True, doc="Bytes representation of private key")
     wif = Column(String(255), index=True, doc="Public or private WIF (Wallet Import Format) representation")
     compressed = Column(Boolean, default=True, doc="Is key compressed or not. Default is True")
     key_type = Column(String(10), default='bip32', doc="Type of key: single, bip32 or multisig. Default is bip32")
@@ -216,16 +236,17 @@ class DbKey(Base):
     is_private = Column(Boolean, doc="Is key private or not?")
     path = Column(String(100), doc="String of BIP-32 key path")
     wallet_id = Column(Integer, ForeignKey('wallets.id'), index=True, doc="Wallet ID which contains this key")
-    wallet = relationship("DbWallet", back_populates="keys", doc="Related HDWallet object")
+    wallet = relationship("DbWallet", back_populates="keys", doc="Related Wallet object")
     transaction_inputs = relationship("DbTransactionInput", cascade="all,delete", back_populates="key",
                                       doc="All DbTransactionInput objects this key is part of")
     transaction_outputs = relationship("DbTransactionOutput", cascade="all,delete", back_populates="key",
                                        doc="All DbTransactionOutput objects this key is part of")
-    balance = Column(Numeric(25, 0, asdecimal=False), default=0, doc="Total balance of UTXO's linked to this key")
+    balance = Column(BigInteger, default=0, doc="Total balance of UTXO's linked to this key")
     used = Column(Boolean, default=False, doc="Has key already been used on the blockchain in as input or output? "
                                               "Default is False")
     network_name = Column(String(20), ForeignKey('networks.name'),
                           doc="Name of key network, i.e. bitcoin, litecoin, dash")
+    latest_txid = Column(LargeBinary(32), doc="TxId of latest transaction downloaded from the blockchain")
     network = relationship("DbNetwork", doc="DbNetwork object for this key")
     multisig_parents = relationship("DbKeyMultisigChildren", backref='child_key',
                                     primaryjoin=id == DbKeyMultisigChildren.child_id,
@@ -234,7 +255,6 @@ class DbKey(Base):
                                      order_by="DbKeyMultisigChildren.key_order",
                                      primaryjoin=id == DbKeyMultisigChildren.parent_id,
                                      doc="List of children keys")
-    latest_txid = Column(String(64), doc="TxId of latest transaction downloaded from the blockchain")
 
     __table_args__ = (
         CheckConstraint(key_type.in_(['single', 'bip32', 'multisig']), name='constraint_key_types_allowed'),
@@ -253,6 +273,9 @@ class DbNetwork(Base):
     """
     Database definitions for networks in Sqlalchemy format
 
+    Most network settings and variables can be found outside the database in the libraries configurations settings.
+    Use the bitcoinlib/data/networks.json file to view and manage settings.
+
     """
     __tablename__ = 'networks'
     name = Column(String(20), unique=True, primary_key=True, doc="Network name, i.e.: bitcoin, litecoin, dash")
@@ -262,12 +285,12 @@ class DbNetwork(Base):
         return "<DbNetwork(name='%s', description='%s'>" % (self.name, self.description)
 
 
-class TransactionType(enum.Enum):
-    """
-    Incoming or Outgoing transaction Enumeration
-    """
-    incoming = 1
-    outgoing = 2
+# class TransactionType(enum.Enum):
+#     """
+#     Incoming or Outgoing transaction Enumeration
+#     """
+#     incoming = 1
+#     outgoing = 2
 
 
 class DbTransaction(Base):
@@ -279,16 +302,17 @@ class DbTransaction(Base):
     """
     __tablename__ = 'transactions'
     id = Column(Integer, Sequence('transaction_id_seq'), primary_key=True,
-                doc="Unique transaction ID for internal usage")
-    hash = Column(String(64), index=True, doc="Hexadecimal representation of transaction hash or transaction ID")
+                doc="Unique transaction index for internal usage")
+    txid = Column(LargeBinary(32), index=True, doc="Bytes representation of transaction ID")
     wallet_id = Column(Integer, ForeignKey('wallets.id'), index=True,
                        doc="ID of wallet which contains this transaction")
+    account_id = Column(Integer, index=True, doc="ID of account")
     wallet = relationship("DbWallet", back_populates="transactions",
-                          doc="Link to HDWallet object which contains this transaction")
+                          doc="Link to Wallet object which contains this transaction")
     witness_type = Column(String(20), default='legacy', doc="Is this a legacy or segwit transaction?")
-    version = Column(Integer, default=1,
+    version = Column(BigInteger, default=1,
                      doc="Tranaction version. Default is 1 but some wallets use another version number")
-    locktime = Column(Integer, default=0,
+    locktime = Column(BigInteger, default=0,
                       doc="Transaction level locktime. Locks the transaction until a specified block "
                           "(value from 1 to 5 million) or until a certain time (Timestamp in seconds after 1-jan-1970)."
                           " Default value is 0 for transactions without locktime")
@@ -300,36 +324,37 @@ class DbTransaction(Base):
                            doc="Number of confirmation when this transaction is included in a block. "
                                "Default is 0: unconfirmed")
     block_height = Column(Integer, index=True, doc="Number of block this transaction is included in")
-    block_hash = Column(String(64), index=True, doc="Transaction is included in block with this hash")
     size = Column(Integer, doc="Size of the raw transaction in bytes")
-    fee = Column(Integer, doc="Transaction fee")
+    fee = Column(BigInteger, doc="Transaction fee")
     inputs = relationship("DbTransactionInput", cascade="all,delete",
                           doc="List of all inputs as DbTransactionInput objects")
     outputs = relationship("DbTransactionOutput", cascade="all,delete",
                            doc="List of all outputs as DbTransactionOutput objects")
     status = Column(String(20), default='new',
-                    doc="Current status of transaction, can be one of the following: new', 'incomplete', "
+                    doc="Current status of transaction, can be one of the following: new', "
                         "'unconfirmed', 'confirmed'. Default is 'new'")
-    input_total = Column(Numeric(25, 0, asdecimal=False), default=0,
+    is_complete = Column(Boolean, default=True, doc="Allow to store incomplete transactions, for instance if not all "
+                                                    "inputs are known when retrieving UTXO's")
+    input_total = Column(BigInteger, default=0,
                          doc="Total value of the inputs of this transaction. Input total = Output total + fee. "
                              "Default is 0")
-    output_total = Column(Numeric(25, 0, asdecimal=False), default=0,
+    output_total = Column(BigInteger, default=0,
                           doc="Total value of the outputs of this transaction. Output total = Input total - fee")
     network_name = Column(String(20), ForeignKey('networks.name'), doc="Blockchain network name of this transaction")
     network = relationship("DbNetwork", doc="Link to DbNetwork object")
-    raw = Column(Text(),
+    raw = Column(LargeBinary,
                  doc="Raw transaction hexadecimal string. Transaction is included in raw format on the blockchain")
     verified = Column(Boolean, default=False, doc="Is transaction verified. Default is False")
 
     __table_args__ = (
-        UniqueConstraint('wallet_id', 'hash', name='constraint_wallet_transaction_hash_unique'),
-        CheckConstraint(status.in_(['new', 'incomplete', 'unconfirmed', 'confirmed']),
+        UniqueConstraint('wallet_id', 'txid', name='constraint_wallet_transaction_hash_unique'),
+        CheckConstraint(status.in_(['new', 'unconfirmed', 'confirmed']),
                         name='constraint_status_allowed'),
         CheckConstraint(witness_type.in_(['legacy', 'segwit']), name='transaction_constraint_allowed_types'),
     )
 
     def __repr__(self):
-        return "<DbTransaction(hash='%s', confirmations='%s')>" % (self.hash, self.confirmations)
+        return "<DbTransaction(txid='%s', confirmations='%s')>" % (self.txid, self.confirmations)
 
 
 class DbTransactionInput(Base):
@@ -347,20 +372,20 @@ class DbTransactionInput(Base):
     key_id = Column(Integer, ForeignKey('keys.id'), index=True, doc="ID of key used in this input")
     key = relationship("DbKey", back_populates="transaction_inputs", doc="Related DbKey object")
     address = Column(String(255),
-                     doc="Address string of input, used if not key is associated. "
-                         "An cryptocurrency address is a hash of the public key")
+                     doc="Address string of input, used if no key is associated. "
+                         "An cryptocurrency address is a hash of the public key or a redeemscript")
     witness_type = Column(String(20), default='legacy',
                           doc="Type of transaction, can be legacy, segwit or p2sh-segwit. Default is legacy")
-    prev_hash = Column(String(64),
+    prev_txid = Column(LargeBinary(32),
                        doc="Transaction hash of previous transaction. Previous unspent outputs (UTXO) is spent "
                            "in this input")
     output_n = Column(BigInteger, doc="Output_n of previous transaction output that is spent in this input")
-    script = Column(Text, doc="Unlocking script to unlock previous locked output")
+    script = Column(LargeBinary, doc="Unlocking script to unlock previous locked output")
     script_type = Column(String(20), default='sig_pubkey',
                          doc="Unlocking script type. Can be 'coinbase', 'sig_pubkey', 'p2sh_multisig', 'signature', "
                              "'unknown', 'p2sh_p2wpkh' or 'p2sh_p2wsh'. Default is sig_pubkey")
     sequence = Column(BigInteger, doc="Transaction sequence number. Used for timelock transaction inputs")
-    value = Column(Numeric(25, 0, asdecimal=False), default=0, doc="Value of transaction input")
+    value = Column(BigInteger, default=0, doc="Value of transaction input")
     double_spend = Column(Boolean, default=False,
                           doc="Indicates if a service provider tagged this transaction as double spend")
 
@@ -386,16 +411,19 @@ class DbTransactionOutput(Base):
                             doc="Transaction ID of parent transaction")
     transaction = relationship("DbTransaction", back_populates='outputs',
                                doc="Link to transaction object")
-    output_n = Column(BigInteger, primary_key=True, doc="Sequence number of transaction output")
+    output_n = Column(Integer, primary_key=True, doc="Sequence number of transaction output")
     key_id = Column(Integer, ForeignKey('keys.id'), index=True, doc="ID of key used in this transaction output")
     key = relationship("DbKey", back_populates="transaction_outputs", doc="List of DbKey object used in this output")
-    script = Column(Text, doc="Locking script which locks transaction output")
+    address = Column(String(255),
+                     doc="Address string of output, used if no key is associated. "
+                         "An cryptocurrency address is a hash of the public key or a redeemscript")
+    script = Column(LargeBinary, doc="Locking script which locks transaction output")
     script_type = Column(String(20), default='p2pkh',
                          doc="Locking script type. Can be one of these values: 'p2pkh', 'multisig', 'p2sh', 'p2pk', "
                              "'nulldata', 'unknown', 'p2wpkh' or 'p2wsh'. Default is p2pkh")
-    value = Column(Numeric(25, 0, asdecimal=False), default=0, doc="Total transaction output value")
-    spent = Column(Boolean(), default=False, doc="Indicated if output is already spent in another transaction")
-    spending_txid = Column(String(64), doc="Transaction hash of input which spends this output")
+    value = Column(BigInteger, default=0, doc="Total transaction output value")
+    spent = Column(Boolean, default=False, doc="Indicated if output is already spent in another transaction")
+    spending_txid = Column(LargeBinary(32), doc="Transaction hash of input which spends this output")
     spending_index_n = Column(Integer, doc="Index number of transaction input which spends this output")
 
     __table_args__ = (CheckConstraint(script_type.in_(['', 'p2pkh',  'multisig', 'p2sh', 'p2pk', 'nulldata',
@@ -413,25 +441,10 @@ def db_update_version_id(db, version):
 
 
 def db_update(db, version_db, code_version=BITCOINLIB_VERSION):
-    if version_db == '0.4.10' and code_version >= '0.4.11':
-        column = Column('latest_txid', String(32))
-        add_column(db.engine, 'keys', column)
-        version_db = db_update_version_id(db, '0.4.11')
-    if version_db == '0.4.11' and code_version >= '0.4.12':
-        column = Column('address', String(255),
-                        doc="Address string of input, used if not key is associated. "
-                            "An cryptocurrency address is a hash of the public key")
-        add_column(db.engine, 'transaction_inputs', column)
-        version_db = db_update_version_id(db, '0.4.12')
-    if version_db <= '0.4.12' and code_version >= '0.4.15':
-        column1 = Column('spending_txid', String(64), doc="Transaction hash of input which spends this output")
-        column2 = Column('spending_index_n', Integer, doc="Index number of transaction input which spends this output")
-        add_column(db.engine, 'transaction_outputs', column1)
-        add_column(db.engine, 'transaction_outputs', column2)
+    # Database changes from version 0.5+
+    #
+    # Older databases cannnot be updated this way, use updatedb.py to copy keys and recreate database.
+    #
 
     version_db = db_update_version_id(db, code_version)
     return version_db
-
-
-if __name__ == '__main__':
-    DbInit()
