@@ -2,7 +2,7 @@
 #
 #    BitcoinLib - Python Cryptocurrency Library
 #    ENCODING - Methods for encoding and conversion
-#    © 2016 - 2020 February - 1200 Web Development <http://1200wd.com/>
+#    © 2016 - 2020 October - 1200 Web Development <http://1200wd.com/>
 #
 #    This program is free software: you can redistribute it and/or modify
 #    it under the terms of the GNU Affero General Public License as
@@ -18,15 +18,12 @@
 #    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 
-import os
 import math
 import numbers
 from copy import deepcopy
 import hashlib
 import pyaes
-import binascii
 import unicodedata
-import struct
 from bitcoinlib.main import *
 _logger = logging.getLogger(__name__)
 
@@ -34,7 +31,7 @@ _logger = logging.getLogger(__name__)
 SCRYPT_ERROR = None
 USING_MODULE_SCRYPT = os.getenv("USING_MODULE_SCRYPT") not in ["false", "False", "0", "FALSE"]
 try:
-    if USING_MODULE_SCRYPT != False:
+    if USING_MODULE_SCRYPT is not False:
         import scrypt
         USING_MODULE_SCRYPT = True
 except ImportError as SCRYPT_ERROR:
@@ -50,7 +47,7 @@ if not USING_MODULE_SCRYPT:
 
 USE_FASTECDSA = os.getenv("USE_FASTECDSA") not in ["false", "False", "0", "FALSE"]
 try:
-    if USE_FASTECDSA != False:
+    if USE_FASTECDSA is not False:
         from fastecdsa.encoding.der import DEREncoder
         USE_FASTECDSA = True
 except ImportError:
@@ -72,7 +69,7 @@ class EncodingError(Exception):
 
 bytesascii = b''
 for bxn in range(256):
-    bytesascii += bytes(bytearray((bxn,)))
+    bytesascii += bytes((bxn,))
 
 code_strings = {
     2: b'01',
@@ -81,7 +78,7 @@ code_strings = {
     16: b'0123456789abcdef',
     32: b'abcdefghijklmnopqrstuvwxyz234567',
     58: b'123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz',
-    256: b''.join([bytes(bytearray((csx,))) for csx in range(256)]),
+    256: b''.join([bytes((csx,)) for csx in range(256)]),
     'bech32': b'qpzry9x8gf2tvdw0s3jn54khce6mua7l'
 }
 
@@ -97,15 +94,12 @@ def _array_to_codestring(array, base):
     codebase = code_strings[base]
     codestring = ""
     for i in array:
-        if not PY3:
-            codestring += codebase[i]
-        else:
-            codestring += chr(codebase[i])
+        codestring += chr(codebase[i])
     return codestring
 
 
 def _codestring_to_array(codestring, base):
-    codestring = to_bytes(codestring)
+    codestring = bytes(codestring, 'utf8')
     codebase = code_strings[base]
     array = []
     for s in codestring:
@@ -125,14 +119,14 @@ def normalize_var(var, base=256):
     Convert decimals to integer type
 
     :param var: input variable in any format
-    :type var: str, byte, bytearray, unicode
+    :type var: str, byte
     :param base: specify variable format, i.e. 10 for decimal, 16 for hex
     :type base: int
 
     :return: Normalized var in string for Python 2, bytes for Python 3, decimal for base10
     """
     try:
-        if PY3 and isinstance(var, str):
+        if isinstance(var, str):
             var = var.encode('ISO-8859-1')
     except ValueError:
         try:
@@ -140,15 +134,6 @@ def normalize_var(var, base=256):
         except ValueError:
             raise EncodingError("Unknown character '%s' in input format" % var)
 
-    if not PY3 and isinstance(var, unicode):
-        try:
-            var = str(var)
-        except UnicodeEncodeError:
-            try:
-                var = var.encode('utf-8')
-            except ValueError:
-                raise EncodingError("Cannot convert this unicode to string format")
-
     if base == 10:
         return int(var)
     elif isinstance(var, list):
@@ -216,26 +201,28 @@ def change_base(chars, base_from, base_to, min_length=0, output_even=None, outpu
     addzeros = 0
     inp = normalize_var(chars, base_from)
 
-    # Use binascii and int for standard conversions to speedup things
+    # Use bytes and int's methods for standard conversions to speedup things
     if not min_length:
         if base_from == 256 and base_to == 16:
-            return to_hexstring(inp)
+            return inp.hex()
         elif base_from == 16 and base_to == 256:
-            return binascii.unhexlify(inp)
-    if base_from == 16 and base_to == 10 and PY3:
+            return bytes.fromhex(chars)
+    if base_from == 16 and base_to == 10:
         return int(inp, 16)
-    if base_from == 10 and base_to == 16 and PY3:
+    if base_from == 10 and base_to == 16:
         hex_outp = hex(inp)[2:]
         return hex_outp.zfill(min_length) if min_length else hex_outp
-    if base_from == 256 and base_to == 10 and PY3:
+    if base_from == 256 and base_to == 10:
         return int.from_bytes(inp, 'big')
+    if base_from == 10 and base_to == 256:
+        return inp.to_bytes(min_length, byteorder='big')
 
     if output_even is None and base_to == 16:
         output_even = True
 
     if isinstance(inp, numbers.Number):
         input_dec = inp
-    elif isinstance(inp, (str, list, bytes, bytearray)):
+    elif isinstance(inp, (str, list, bytes)):
         factor = 1
         while len(inp):
             if isinstance(inp, list):
@@ -253,11 +240,8 @@ def change_base(chars, base_from, base_to, min_length=0, output_even=None, outpu
             input_dec += pos * factor
 
             # Add leading zero if there are leading zero's in input
+            firstchar = chr(code_str_from[0]).encode('utf-8')
             if not pos * factor:
-                if not PY3:
-                    firstchar = code_str_from[0]
-                else:
-                    firstchar = chr(code_str_from[0]).encode('utf-8')
                 if isinstance(inp, list):
                     if not len([x for x in inp if x != firstchar]):
                         addzeros += 1
@@ -275,9 +259,17 @@ def change_base(chars, base_from, base_to, min_length=0, output_even=None, outpu
     if base_to != 10:
         pos_fact = math.log(base_to, base_from)
         expected_length = len(str(chars)) / pos_fact
+
         zeros = int(addzeros / pos_fact)
         if addzeros == 1:
             zeros = 1
+        # Different rules for base58 addresses
+        if (base_from == 256 and base_to == 58) or (base_from == 58 and base_to == 256):
+            zeros = addzeros
+        elif base_from == 16 and base_to == 58:
+            zeros = -(-addzeros // 2)
+        elif base_from == 58 and base_to == 16:
+            zeros = addzeros * 2
 
         for _ in range(zeros):
             if base_to != 10 and not expected_length == len(output):
@@ -294,25 +286,14 @@ def change_base(chars, base_from, base_to, min_length=0, output_even=None, outpu
     if not output_as_list and isinstance(output, list):
         if len(output) == 0:
             output = 0
-        elif not PY3:
-            output = ''.join(output)
         else:
             co = ''
             for c in output:
                 co += chr(c)
             output = co
-        # elif isinstance(output[0], bytes):
-        #     output = b''.join(output)
-        # elif isinstance(output[0], int):
-        #     co = ''
-        #     for c in output:
-        #         co += chr(c)
-        #     output = co
-        # else:
-        #     output = ''.join(output)
     if base_to == 10:
         return int(0) or (output != '' and int(output))
-    if PY3 and base_to == 256 and not output_as_list:
+    if base_to == 256 and not output_as_list:
         return output.encode('ISO-8859-1')
     else:
         return output
@@ -324,20 +305,19 @@ def varbyteint_to_int(byteint):
 
     See https://en.bitcoin.it/wiki/Protocol_documentation#Variable_length_integer for specification
 
-    >>> varbyteint_to_int(to_bytes('fd1027'))
+    >>> varbyteint_to_int(bytes.fromhex('fd1027'))
     (10000, 3)
 
     :param byteint: 1-9 byte representation
-    :type byteint: bytes, list, bytearray
+    :type byteint: bytes, list
 
     :return (int, int): tuple wit converted integer and size
     """
-    if not isinstance(byteint, (bytes, list, bytearray)):
+    if not isinstance(byteint, (bytes, list)):
         raise EncodingError("Byteint must be a list or defined as bytes")
-    if PY3 or isinstance(byteint, (list, bytearray)):
-        ni = byteint[0]
-    else:
-        ni = ord(byteint[0])
+    if byteint == b'':
+        return 0
+    ni = byteint[0]
     if ni < 253:
         return ni, 1
     if ni == 253:  # integer of 2 bytes
@@ -346,7 +326,22 @@ def varbyteint_to_int(byteint):
         size = 4
     else:  # integer of 8 bytes
         size = 8
-    return change_base(byteint[1:1+size][::-1], 256, 10), size + 1
+    return int.from_bytes(byteint[1:1+size][::-1], 'big'), size + 1
+
+
+def read_varbyteint(s):
+    """
+    Read variable length integer from BytesIO stream. Wrapper for the varbyteint_to_int method
+
+    :param s: A binary stream
+    :type s: BytesIO
+
+    :return int:
+    """
+    pos = s.tell()
+    value, size = varbyteint_to_int(s.read(9))
+    s.seek(pos + size)
+    return value
 
 
 def int_to_varbyteint(inp):
@@ -355,7 +350,7 @@ def int_to_varbyteint(inp):
 
     See https://en.bitcoin.it/wiki/Protocol_documentation#Variable_length_integer for specification
 
-    >>> to_hexstring(int_to_varbyteint(10000))
+    >>> int_to_varbyteint(10000).hex()
     'fd1027'
 
     :param inp: Integer to convert
@@ -366,13 +361,13 @@ def int_to_varbyteint(inp):
     if not isinstance(inp, numbers.Number):
         raise EncodingError("Input must be a number type")
     if inp < 0xfd:
-        return struct.pack('B', inp)
+        return inp.to_bytes(1, 'little')
     elif inp < 0xffff:
-        return struct.pack('<cH', b'\xfd', inp)
+        return b'\xfd' + inp.to_bytes(2, 'little')
     elif inp < 0xffffffff:
-        return struct.pack('<cL', b'\xfe', inp)
+        return b'\xfe' + inp.to_bytes(4, 'little')
     else:
-        return struct.pack('<cQ', b'\xff', inp)
+        return b'\xff' + inp.to_bytes(8, 'little')
 
 
 def convert_der_sig(signature, as_hex=True):
@@ -401,7 +396,7 @@ def convert_der_sig(signature, as_hex=True):
     if as_hex:
         return sig
     else:
-        return binascii.unhexlify(sig)
+        return bytes.fromhex(sig)
 
 
 def der_encode_sig(r, s):
@@ -469,12 +464,14 @@ def addr_base58_to_pubkeyhash(address, as_hex=False):
         address = change_base(address, 58, 256, 25)
     except EncodingError as err:
         raise EncodingError("Invalid address %s: %s" % (address, err))
+    if len(address) != 25:
+        raise EncodingError("Invalid address hash160 length, should be 25 characters not %d" % len(address))
     check = address[-4:]
     pkh = address[:-4]
     checksum = double_sha256(pkh)[0:4]
     assert (check == checksum), "Invalid address, checksum incorrect"
     if as_hex:
-        return change_base(pkh, 256, 16)[2:]
+        return pkh.hex()[2:]
     else:
         return pkh[1:]
 
@@ -514,7 +511,7 @@ def addr_bech32_to_pubkeyhash(bech, prefix=None, include_witver=False, as_hex=Fa
     if not _bech32_polymod(hrp_expanded + data) == 1:
         raise EncodingError("Bech polymod check failed")
     data = data[:-6]
-    decoded = bytearray(convertbits(data[1:], 5, 8, pad=False))
+    decoded = bytes(convertbits(data[1:], 5, 8, pad=False))
     if decoded is None or len(decoded) < 2 or len(decoded) > 40:
         raise EncodingError("Invalid decoded data length, must be between 2 and 40")
     if data[0] > 16:
@@ -524,9 +521,9 @@ def addr_bech32_to_pubkeyhash(bech, prefix=None, include_witver=False, as_hex=Fa
     prefix = b''
     if include_witver:
         datalen = len(decoded)
-        prefix = bytearray([data[0] + 0x50 if data[0] else 0, datalen])
+        prefix = bytes([data[0] + 0x50 if data[0] else 0, datalen])
     if as_hex:
-        return change_base(prefix + decoded, 256, 16)
+        return (prefix + decoded).hex()
     return prefix + decoded
 
 
@@ -572,8 +569,7 @@ def pubkeyhash_to_addr_base58(pubkeyhash, prefix=b'\x00'):
 
     :return str: Base-58 encoded address
     """
-    # prefix = to_bytes(prefix)
-    key = to_bytearray(prefix) + to_bytearray(pubkeyhash)
+    key = to_bytes(prefix) + to_bytes(pubkeyhash)
     addr256 = key + double_sha256(key)[:4]
     return change_base(addr256, 256, 58)
 
@@ -590,7 +586,7 @@ def pubkeyhash_to_addr_bech32(pubkeyhash, prefix='bc', witver=0, separator='1'):
     For more information see BIP173 proposal at https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki
 
     :param pubkeyhash: Public key hash
-    :type pubkeyhash: str, bytes, bytearray
+    :type pubkeyhash: str, bytes
     :param prefix: Address prefix or Human-readable part. Default is 'bc' an abbreviation of Bitcoin. Use 'tb' for testnet.
     :type prefix: str
     :param witver: Witness version between 0 and 16
@@ -601,8 +597,7 @@ def pubkeyhash_to_addr_bech32(pubkeyhash, prefix='bc', witver=0, separator='1'):
     :return str: Bech32 encoded address
     """
 
-    if not isinstance(pubkeyhash, bytearray):
-        pubkeyhash = bytearray(to_bytes(pubkeyhash))
+    pubkeyhash = list(to_bytes(pubkeyhash))
 
     if len(pubkeyhash) not in [20, 32]:
         if int(pubkeyhash[0]) != 0:
@@ -640,7 +635,7 @@ def convertbits(data, frombits, tobits, pad=True):
     Source: https://github.com/sipa/bech32/tree/master/ref/python
 
     :param data: Data values to convert
-    :type data: list, bytearray
+    :type data: list
     :param frombits: Number of bits in source data
     :type frombits: int
     :param tobits: Number of bits in result data
@@ -656,8 +651,6 @@ def convertbits(data, frombits, tobits, pad=True):
     maxv = (1 << tobits) - 1
     max_acc = (1 << (frombits + tobits - 1)) - 1
     for value in data:
-        if not PY3 and isinstance(value, str):
-            value = int(value, 16)
         if value < 0 or (value >> frombits):
             return None
         acc = ((acc << frombits) | value) & max_acc
@@ -677,7 +670,7 @@ def varstr(string):
     """
     Convert string to variably sized string: Bytestring preceded with length byte
 
-    >>> to_hexstring(varstr(to_bytes('5468697320737472696e67206861732061206c656e677468206f66203330')))
+    >>> varstr(to_bytes('5468697320737472696e67206861732061206c656e677468206f66203330')).hex()
     '1e5468697320737472696e67206861732061206c656e677468206f66203330'
 
     :param string: String input
@@ -691,73 +684,57 @@ def varstr(string):
     return int_to_varbyteint(len(s)) + s
 
 
-def to_bytearray(string):
-    """
-    Convert String, Unicode or Bytes to Python 2 and 3 compatible ByteArray
-
-    :param string: String, Unicode, Bytes or ByteArray
-    :type string: bytes, str, bytearray
-
-    :return bytearray:
-    """
-    if isinstance(string, TYPE_TEXT):
-        try:
-            string = binascii.unhexlify(string)
-        except (TypeError, binascii.Error):
-            pass
-    return bytearray(string)
-
-
 def to_bytes(string, unhexlify=True):
     """
-    Convert String, Unicode or ByteArray to Bytes
+    Convert string, hexadecimal string  to bytes
 
     :param string: String to convert
-    :type string: str, unicode, bytes, bytearray
+    :type string: str, bytes
     :param unhexlify: Try to unhexlify hexstring
     :type unhexlify: bool
 
     :return: Bytes var
     """
-    s = normalize_var(string)
+    if not string:
+        return b''
     if unhexlify:
         try:
-            s = binascii.unhexlify(s)
+            if isinstance(string, bytes):
+                string = string.decode()
+            s = bytes.fromhex(string)
             return s
-        except (TypeError, binascii.Error):
+        except (TypeError, ValueError):
             pass
-    return s
+    if isinstance(string, bytes):
+        return string
+    else:
+        return bytes(string, 'utf8')
 
 
 def to_hexstring(string):
     """
-    Convert Bytes or ByteArray to hexadecimal string
+    Convert bytes, string to a hexadecimal string. Use instead of built-in hex() method if format
+    of input string is not known.
 
-    >>> to_hexstring('\x12\xaa\xdd')
+    >>> to_hexstring(b'\\x12\\xaa\\xdd')
     '12aadd'
 
     :param string: Variable to convert to hex string
-    :type string: bytes, bytearray, str
+    :type string: bytes, str
 
     :return: hexstring
     """
-    string = normalize_var(string)
+    if not string:
+        return ''
+    try:
+        bytes.fromhex(string)
+        return string
+    except (ValueError, TypeError):
+        pass
 
-    if isinstance(string, (str, bytes)):
-        try:
-            binascii.unhexlify(string)
-            if PY3:
-                return str(string, 'ISO-8859-1')
-            else:
-                return string
-        except (TypeError, binascii.Error):
-            pass
-
-    s = binascii.hexlify(string)
-    if PY3:
-        return str(s, 'ISO-8859-1')
-    else:
-        return s
+    if not isinstance(string, bytes):
+        string = bytes(string, 'utf8')
+    return string.hex()
 
 
 def normalize_string(string):
@@ -766,11 +743,11 @@ def normalize_string(string):
     See https://en.wikipedia.org/wiki/Unicode_equivalence#Normalization
 
     :param string: string value
-    :type string: bytes, bytearray, str
+    :type string: bytes, str
 
     :return: string
     """
-    if isinstance(string, str if sys.version < '3' else bytes):
+    if isinstance(string, bytes):
         utxt = string.decode('utf8')
     elif isinstance(string, TYPE_TEXT):
         utxt = string
@@ -809,7 +786,7 @@ def hash160(string):
     return hashlib.new('ripemd160', hashlib.sha256(string).digest()).digest()
 
 
-def bip38_decrypt(encrypted_privkey, passphrase):
+def bip38_decrypt(encrypted_privkey, password):
     """
     BIP0038 non-ec-multiply decryption. Returns WIF private key.
     Based on code from https://github.com/nomorecoin/python-bip38-testing
@@ -817,8 +794,8 @@ def bip38_decrypt(encrypted_privkey, passphrase):
 
     :param encrypted_privkey: Encrypted private key using WIF protected key format
     :type encrypted_privkey: str
-    :param passphrase: Required passphrase for decryption
-    :type passphrase: str
+    :param password: Required password for decryption
+    :type password: str
 
     :return tupple (bytes, bytes): (Private Key bytes, 4 byte address hash for verification)
     """
@@ -831,11 +808,11 @@ def bip38_decrypt(encrypted_privkey, passphrase):
         compressed = True
     else:
         raise EncodingError("Unrecognised password protected key format. Flagbyte incorrect.")
-    if isinstance(passphrase, str) and sys.version_info > (3,):
-        passphrase = passphrase.encode('utf-8')
+    if isinstance(password, str):
+        password = password.encode('utf-8')
     addresshash = d[0:4]
     d = d[4:-4]
-    key = scrypt.hash(passphrase, addresshash, 16384, 8, 8, 64)
+    key = scrypt.hash(password, addresshash, 16384, 8, 8, 64)
     derivedhalf1 = key[0:32]
     derivedhalf2 = key[32:64]
     encryptedhalf1 = d[0:16]
@@ -844,14 +821,14 @@ def bip38_decrypt(encrypted_privkey, passphrase):
     decryptedhalf2 = aes.decrypt(encryptedhalf2)
     decryptedhalf1 = aes.decrypt(encryptedhalf1)
     priv = decryptedhalf1 + decryptedhalf2
-    priv = binascii.unhexlify('%064x' % (int(binascii.hexlify(priv), 16) ^ int(binascii.hexlify(derivedhalf1), 16)))
+    priv = (int.from_bytes(priv, 'big') ^ int.from_bytes(derivedhalf1, 'big')).to_bytes(32, 'big')
     # if compressed:
     #     # FIXME: This works but does probably not follow the BIP38 standards (was before: priv = b'\0' + priv)
     #     priv += b'\1'
     return priv, addresshash, compressed
 
 
-def bip38_encrypt(private_hex, address, passphrase, flagbyte=b'\xe0'):
+def bip38_encrypt(private_hex, address, password, flagbyte=b'\xe0'):
     """
     BIP0038 non-ec-multiply encryption. Returns BIP0038 encrypted private key
     Based on code from https://github.com/nomorecoin/python-bip38-testing
@@ -860,26 +837,26 @@ def bip38_encrypt(private_hex, address, passphrase, flagbyte=b'\xe0'):
     :type private_hex: str
     :param address: Address string
     :type address: str
-    :param passphrase: Required passphrase for encryption
-    :type passphrase: str
+    :param password: Required password for encryption
+    :type password: str
     :param flagbyte: Flagbyte prefix for WIF
     :type flagbyte: bytes
 
-    :return str: BIP38 passphrase encrypted private key
+    :return str: BIP38 password encrypted private key
     """
-    if isinstance(address, str) and sys.version_info > (3,):
+    if isinstance(address, str):
         address = address.encode('utf-8')
-    if isinstance(passphrase, str) and sys.version_info > (3,):
-        passphrase = passphrase.encode('utf-8')
+    if isinstance(password, str):
+        password = password.encode('utf-8')
     addresshash = double_sha256(address)[0:4]
-    key = scrypt.hash(passphrase, addresshash, 16384, 8, 8, 64)
+    key = scrypt.hash(password, addresshash, 16384, 8, 8, 64)
     derivedhalf1 = key[0:32]
     derivedhalf2 = key[32:64]
     aes = pyaes.AESModeOfOperationECB(derivedhalf2)
-    encryptedhalf1 = aes.encrypt(binascii.unhexlify('%0.32x' % (int(private_hex[0:32], 16) ^
-                                                                int(binascii.hexlify(derivedhalf1[0:16]), 16))))
-    encryptedhalf2 = aes.encrypt(binascii.unhexlify('%0.32x' % (int(private_hex[32:64], 16) ^
-                                                                int(binascii.hexlify(derivedhalf1[16:32]), 16))))
+    encryptedhalf1 = \
+        aes.encrypt((int(private_hex[0:32], 16) ^ int.from_bytes(derivedhalf1[0:16], 'big')).to_bytes(16, 'big'))
+    encryptedhalf2 = \
+        aes.encrypt((int(private_hex[32:64], 16) ^ int.from_bytes(derivedhalf1[16:32], 'big')).to_bytes(16, 'big'))
     encrypted_privkey = b'\x01\x42' + flagbyte + addresshash + encryptedhalf1 + encryptedhalf2
     encrypted_privkey += double_sha256(encrypted_privkey)[:4]
     return change_base(encrypted_privkey, 256, 58)
