@@ -21,13 +21,12 @@
 import os
 import unittest
 import json
-
 from bitcoinlib.networks import NETWORK_DEFINITIONS
 from bitcoinlib.keys import *
 
 # Number of bulktests for generation of private, public keys and HDKeys. Set to 0 to disable
 # WARNING: Can be slow for a larger number of tests
-BULKTESTCOUNT = 10
+BULKTESTCOUNT = 100
 
 
 class TestKeyClasses(unittest.TestCase):
@@ -36,14 +35,19 @@ class TestKeyClasses(unittest.TestCase):
         pk = 'xprv9s21ZrQH143K4EDmQNMBqXwUTcrRoUctKkTegGsaBcMLnR1fJkMjVSRwVswjHzJspfWCUwzge1F521cY4wfWD54tzXVUqeo' \
              'TFkZo17HiK2y'
         k = HDKey(pk)
-        self.assertEqual(str(k), 'd3caaca97fbba3d4ebd87e855c6e05080b0ac9118bd886c0f575d79940bd6eb4')
+        self.assertEqual(str(k), '03dc86716b2be27a0575558bac73279290ac22c3ea0240e42a2152d584f2b4006b')
+        self.assertEqual(len(k), 33)
         self.assertEqual(int(k), 95796105828208927954168018443072630832764875640480247096632116413925408206516)
         k2 = HDKey(pk)
         self.assertTrue(k == k2)
         pubk2 = HDKey(k.wif_public())
         self.assertEqual(str(pubk2), '03dc86716b2be27a0575558bac73279290ac22c3ea0240e42a2152d584f2b4006b')
         self.assertTrue(k.public() == pubk2)
-        
+        self.assertEqual(k + k2, b'\x03\xdc\x86qk+\xe2z\x05uU\x8b\xacs\'\x92\x90\xac"\xc3\xea\x02@\xe4*!R\xd5'
+                                 b'\x84\xf2\xb4\x00k\x03\xdc\x86qk+\xe2z\x05uU\x8b\xacs\'\x92\x90\xac"'
+                                 b'\xc3\xea\x02@\xe4*!R\xd5\x84\xf2\xb4\x00k')
+        self.assertEqual(k + k2, k.public_byte + k2.public_byte)
+
     def test_dict_and_json_outputs(self):
         k = HDKey()
         k.address(script_type='p2wsh', encoding='bech32')
@@ -129,8 +133,6 @@ class TestPrivateKeyConversions(unittest.TestCase):
 
     def test_private_key_conversions_wif(self):
         self.assertEqual('L3RyKcjp8kzdJ6rhGhTC5bXWEYnC2eL3b1vrZoduXMht6m9MQeHy', self.k.wif())
-        if not PY3:
-            return
         self.assertEqual('XHVtmt8BSSd5MRs5JTT4apiX9a3mUSwHxbGm6Ky6qiyyVvFRhmU7', self.k.wif(prefix='cc'))
         self.assertEqual('XHVtmt8BSSd5MRs5JTT4apiX9a3mUSwHxbGm6Ky6qiyyVvFRhmU7', self.k.wif(prefix=b'\xcc'))
 
@@ -165,13 +167,7 @@ class TestPrivateKeyImport(unittest.TestCase):
         pk = '4781e448a7ff0e1b66f1a249b4c952dae33326cf57c0a643738886f4efcd14d57a380bc32c26f46e733cd' \
              '991064c2e7f7d532b9c9ca825671a8809ab6876c78b'
         k = Key(pk)
-        self.assertEqual('f93677c417d4750c7a5806f849739265cc46b8a9', to_hexstring(k.hash160))
-
-    def test_private_key_import_key_bytearray(self):
-        pk = bytearray(b':\xbaAb\xc7%\x1c\x89\x12\x07\xb7G\x84\x05Q\xa7\x199\xb0\xde\x08\x1f\x85\xc4\xe4L\xf7\xc1>'
-                       b'A\xda\xa6\x01')
-        self.k = Key(pk)
-        self.assertEqual('KyBsPXxTuVD82av65KZkrGrWi5qLMah5SdNq6uftawDbgKa2wv6S', self.k.wif())
+        self.assertEqual('f93677c417d4750c7a5806f849739265cc46b8a9', k.hash160.hex())
 
     def test_private_key_import_wif(self):
         self.k = Key('L1odb1uUozbfK2NrsMyhJfvRsxGM2AxixgPL8vG9BUBnE6W1VyTX')
@@ -204,27 +200,27 @@ class TestPrivateKeyImport(unittest.TestCase):
 class TestPublicKeyConversion(unittest.TestCase):
 
     def setUp(self):
-        self.publickey_hex = '044781e448a7ff0e1b66f1a249b4c952dae33326cf57c0a643738886f4efcd14d57a380bc32c26f46e733cd' \
-                             '991064c2e7f7d532b9c9ca825671a8809ab6876c78b'
+        self.publickey_hex = '043be860d524b1aef015d0e501333d5e62fd39652bfb89e6720c3eb1cb10754370eeee80b04abf2f80dec' \
+                             '69431498723f6411e8e03446796fa250f8dfe3fa8ff84'
         self.K = Key(self.publickey_hex)
         self.KC = Key('034781e448a7ff0e1b66f1a249b4c952dae33326cf57c0a643738886f4efcd14d5')
 
     def test_public_key_get_address_uncompressed(self):
-        self.assertEqual('12ooWDQp6mujkVpEWHdfHmfM4rU17bokWw', self.K.address_uncompressed())
+        self.assertEqual('1BwbZw1fG91jMYsXh6hfnvPGXBaMcughNL', self.K.address_uncompressed())
 
     def test_public_key_get_address(self):
         self.assertEqual('1P2X35YnajqoBXtPpQXJzV1QMnqSZQsn82', self.KC.address())
 
     def test_public_key_get_point(self):
-        self.assertEqual((32343711077743629729728681292399790965391040816412086995020432364076041835733,
-                          55281192143835269607479311758661973079027103826274522268778194868406595274635),
+        self.assertEqual((27097034899423571266687886742220335326047800315940073646681141977460289913712,
+                          108071855740589417275074236385664540865647915362450545389459678127463347060612),
                          self.K.public_point())
 
     def test_public_key_get_hash160_uncompressed(self):
-        self.assertEqual('13d21450578cd8f8645d2e56e684deb7cd77864b', to_hexstring(self.K.hash160))
+        self.assertEqual('78049354383f043fb15a04be58a289ef8a2c03fa', self.K.hash160.hex())
 
     def test_public_key_get_hash160(self):
-        self.assertEqual('f19c417fd97e364afb06e1edd2c0e6a7ecf1af00', to_hexstring(self.KC.hash160))
+        self.assertEqual('f19c417fd97e364afb06e1edd2c0e6a7ecf1af00', self.KC.hash160.hex())
 
     def test_public_key_try_private(self):
         self.assertFalse(self.K.private_hex)
@@ -309,7 +305,7 @@ class TestHDKeysImport(unittest.TestCase):
     def test_hdkey_import_bip38_key(self):
         if USING_MODULE_SCRYPT:
             self.k = HDKey('6PYNKZ1EAgYgmQfmNVamxyXVWHzK5s6DGhwP4J5o44cvXdoY7sRzhtpUeo',
-                           passphrase='TestingOneTwoThree')
+                           password='TestingOneTwoThree')
             self.assertEqual('L44B5gGEpqEDRS9vVPz7QT35jcBG2r3CZwSwQ4fCewXAhAhqGVpP', self.k.wif_key())
 
     def test_hdkey_import_public(self):
@@ -342,8 +338,6 @@ class TestHDKeysImport(unittest.TestCase):
             self.assertEqual(HDKey(wif[0]).wif(is_private=wif[1]), wif[0])
 
     def test_hdkey_import_from_private_byte(self):
-        if not PY3:
-            return
         keystr = b"fch\xe4w\xa8\xdd\xd4h\x08\xc5'\xcc<Pg\x19\xbb?R\xa9'\xb6\xc152\x98KqKV\xad\x91`G-a\xb1\xad\xd8eL" \
                  b"\xcc\x8an\x94\xa3\x93\xb5\xa5\xe6\xc3\xf1\x98\x91h6wt\xf0z=\x1f\x17"
         hdkey = HDKey(keystr)
@@ -538,7 +532,7 @@ class TestBip38(unittest.TestCase):
         if not USING_MODULE_SCRYPT:
             return
         for v in self.vectors["valid"]:
-            k = Key(v['bip38'], passphrase=str(v['passphrase']))
+            k = Key(v['bip38'], password=str(v['passphrase']))
             print("Check %s - %s = %s " % (v['bip38'], v['passphrase'], v['wif']))
             self.assertEqual(str(v['wif']), k.wif())
 
@@ -556,7 +550,7 @@ class TestBip38(unittest.TestCase):
         for network in networks:
             k = Key(network=network)
             enc_key = k.bip38_encrypt('password')
-            k2 = Key(enc_key, passphrase='password', network=network)
+            k2 = Key(enc_key, password='password', network=network)
             self.assertEqual(k.wif(), k2.wif())
 
     def test_bip38_hdkey_method(self):
@@ -695,7 +689,7 @@ class TestKeysAddress(unittest.TestCase):
     def test_keys_hdkey_segwit(self):
         k1 = HDKey('L1TZxZ9RgwFKiGPm6P7J9REQFKG9ymwLSsTwQSwxzLyDJs3CcRkF', witness_type='segwit')
         self.assertEqual(k1.address(), 'bc1qmk9myu4zf590ae2mfq3m63rlfhd5scatl4ckmw')
-
+        self.assertEqual(k1.address_obj.data, '024429dd84f7c12c83f1920d60c7edb60c61d03fa5b8a8b526f4608ae9af89d9f3')
         phrase = 'scan display embark segment deputy lesson vanish second wonder erase crumble swing'
         k2 = HDKey.from_passphrase(phrase, witness_type='segwit', multisig=True)
         self.assertEqual(k2.address(), 'bc1qvj6c7n0hpl9t5r80zya4uukf0zens8ulxgwc0avnxsengtr5juss4pqeqy')
@@ -728,9 +722,9 @@ class TestKeysDash(unittest.TestCase):
         pk = 'xprv9s21ZrQH143K3cq8ueA8GV9uv7cHqkyQGBQu8YZkAU2EXG5oSKVFeQnYK25zhHEEqqjfyTFEcV5enh6vh4tFA3FvdGuWAqPqvY' \
              'ECNLB78mV'
         k = HDKey(pk, network='dash')
-        self.assertEqual(k.account_key().wif(is_private=True), 'xprv9ySHTHmm4KdkKa2RV2zuSmVUAPNynEvkrCDVa95Js9StLECY2'
-                                                               'RjuxNpHKaVfA2hnjob5Zumx1kTg3MhQPsZf7W5h8aEM61AMSqz1zV'
-                                                               'Wjt4Q')
+        self.assertEqual(k.public_master().wif(),
+                         'xpub6CRdroJethC3Y46tb4XuouSCiRDUBhecDR96NXUvRUysD2XgZy4AWB8mAsvMmcw9GgXvmu4BRSFj1yAdiN1K7f'
+                         'w9o96T41hLJRLpLGLJrxY')
 
     def test_hdkey_dash(self):
         k = HDKey('xprv9s21ZrQH143K4EGnYMHVxNp8JgqXCyywC3CGTrSzSudH3iRgC1gPTYgce4xamXMnyDAX8Qv8tvuW1LEgkZSrXiC25LqTJN'
@@ -746,7 +740,7 @@ class TestKeysSignatures(unittest.TestCase):
 
     def test_signatures(self):
         sig_tests = [
-            # tx_hash, key_hex, k, signature, DER encoded sign.
+            # txid, key_hex, k, signature, DER encoded sign.
             ('0d12fdc4aac9eaaab9730999e0ce84c3bd5bb38dfd1f4c90c613ee177987429c',
              'b2da575054fb5daba0efde613b0b8e37159b8110e4be50f73cbe6479f6038f5b', 1002,
              '70b55404702ffa86ecfa4e88e0f354004a0965a5eea5fbbd297436001ae920df5da0917d7bd645c2a09671894375e3d353313'
@@ -786,18 +780,18 @@ class TestKeysSignatures(unittest.TestCase):
         ]
         sig_method1 = sign(sig_tests[0][0], sig_tests[0][1], k=sig_tests[0][2])
         self.assertEqual(sig_method1.hex(), sig_tests[0][3])
-        self.assertEqual(to_hexstring(sig_method1.as_der_encoded()), sig_tests[0][4])
+        self.assertEqual(sig_method1.as_der_encoded(include_hash_type=False).hex(), sig_tests[0][4])
         count = 0
         for case in sig_tests:
             sig = Signature.create(case[0], case[1], k=case[2])
             self.assertEqual(sig.hex(), case[3], msg="Error in #%d: %s != %s" % (count, sig.hex(), case[3]))
-            self.assertEqual(to_hexstring(sig.as_der_encoded()), case[4])
+            self.assertEqual(sig.as_der_encoded(include_hash_type=False).hex(), case[4])
             self.assertTrue(sig.verify())
             count += 1
 
-    def test_rfc6979(self):
+    def test_signatures_rfc6979(self):
         if not USE_FASTECDSA:
-            # This test are only usefull when fastecdsa library is used
+            # This test are only useful when fastecdsa library is used
             return True
 
         # source: https://bitcointalk.org/index.php?topic=285142.40
@@ -870,11 +864,11 @@ class TestKeysSignatures(unittest.TestCase):
             sig = sign(msg_hash, x, k=k)
             self.assertEqual(sig.hex(), vector[3])
 
-    def test_sig_from_r_and_s(self):
+    def test_signatures_from_r_and_s(self):
         r = 0x657912a72d3ac8169fe8eaecd5ab401c94fc9981717e3e6dd4971889f785790c
         s = 0x00ed3bf3456eb76677fd899c8ccd1cc6d1ebc631b94c42f7c4578f28590d651c6e
         expected_der = '30450220657912a72d3ac8169fe8eaecd5ab401c94fc9981717e3e6dd4971889f785790c022100ed3bf345' \
-                       '6eb76677fd899c8ccd1cc6d1ebc631b94c42f7c4578f28590d651c6e'
+                       '6eb76677fd899c8ccd1cc6d1ebc631b94c42f7c4578f28590d651c6e01'
         expected_sig_bytes = b'ey\x12\xa7-:\xc8\x16\x9f\xe8\xea\xec\xd5\xab@\x1c\x94\xfc\x99\x81q~>m\xd4\x97\x18' \
                              b'\x89\xf7\x85y\x0c\xed;\xf3En\xb7fw\xfd\x89\x9c\x8c\xcd\x1c\xc6\xd1\xeb\xc61\xb9LB' \
                              b'\xf7\xc4W\x8f(Y\re\x1cn'
@@ -882,11 +876,11 @@ class TestKeysSignatures(unittest.TestCase):
                            'c8ccd1cc6d1ebc631b94c42f7c4578f28590d651c6e'
 
         sig = Signature(r, s)
-        self.assertEqual(to_hexstring(sig.as_der_encoded()), expected_der)
+        self.assertEqual(sig.as_der_encoded().hex(), expected_der)
         self.assertEqual(sig.bytes(), expected_sig_bytes)
         self.assertEqual(sig.hex(), expected_sig_hex)
 
-    def test_sig_rs_out_of_curve(self):
+    def test_signatures_rs_out_of_curve(self):
         outofcurveint = 115792089237316195423570985008687907852837564279074904382605163141518161494339
         self.assertRaisesRegexp(BKeyError, "r is not a positive integer smaller than the curve order",
                                 Signature, outofcurveint, 10)
@@ -895,6 +889,26 @@ class TestKeysSignatures(unittest.TestCase):
         self.assertRaisesRegexp(BKeyError, "s is not a positive integer smaller than the curve order",
                                 Signature, 11, outofcurveint)
 
+    def test_signatures_dunder(self):
+        sig1 = Signature.parse_hex('3045022100c949a465a057f3ca7d20e80511e93d0be21e3efbeb8720ca3e0adfbce6883d0a022070b'
+                                   '2c6bee101a4ffcb854bae34dbd1f35c31140a46559148a1fa883eedede03401')
+        sig2 = Signature.parse_hex('3045022100b5ce13dc408c65208cf475b44b2012845d4d3fb7a2cacfa35f6b5143761f976f02207d8'
+                                   '581d6004779c7f168e90496d544407d5f0e2eecd44c50fcef1006a86731ec01')
+        self.assertEqual(len(sig1), 72)
+        self.assertEqual(sig1 + sig2, sig1.as_der_encoded() + sig2.as_der_encoded())
+        self.assertEqual(sig1 + sig2, b'0E\x02!\x00\xc9I\xa4e\xa0W\xf3\xca} \xe8\x05\x11\xe9=\x0b\xe2\x1e>'
+                                      b'\xfb\xeb\x87 \xca>\n\xdf\xbc\xe6\x88=\n\x02 p\xb2\xc6\xbe\xe1'
+                                      b'\x01\xa4\xff\xcb\x85K\xae4\xdb\xd1\xf3\\1\x14\nFU\x91H\xa1\xfa\x88>\xed'
+                                      b'\xed\xe04\x010E\x02!\x00\xb5\xce\x13\xdc@\x8ce \x8c\xf4u\xb4K \x12\x84]M?'
+                                      b'\xb7\xa2\xca\xcf\xa3_kQCv\x1f\x97o\x02 }\x85\x81\xd6\x00Gy\xc7\xf1'
+                                      b'h\xe9\x04\x96\xd5D@}_\x0e.\xec\xd4LP\xfc\xef\x10\x06\xa8g1\xec\x01')
+        self.assertEqual(str(sig1), '3045022100c949a465a057f3ca7d20e80511e93d0be21e3efbeb8720ca3e0adfbce6883d0a0220'
+                                    '70b2c6bee101a4ffcb854bae34dbd1f35c31140a46559148a1fa883eedede03401')
+        self.assertEqual(bytes(sig1), b'0E\x02!\x00\xc9I\xa4e\xa0W\xf3\xca} \xe8\x05\x11\xe9=\x0b\xe2\x1e>'
+                                      b'\xfb\xeb\x87 \xca>\n\xdf\xbc\xe6\x88=\n\x02 p\xb2\xc6\xbe\xe1'
+                                      b'\x01\xa4\xff\xcb\x85K\xae4\xdb\xd1\xf3\\1\x14\nFU\x91H\xa1\xfa\x88>\xed'
+                                      b'\xed\xe04\x01')
+
 
 if __name__ == '__main__':
     unittest.main()
